@@ -62,6 +62,7 @@ class PM {
     if($passwort != ""){
       $code=$this->code.$passwort;
       $PW=1;
+      $this->update_pwcounter();
     } else {
       $code=$this->code;
       $PW=0;
@@ -332,6 +333,19 @@ class PM {
       // echo $this->meldung;
     }
 
+  private function update_pwcounter() {
+    $neuerwert=0;
+    $wert = $this->pwcounter();
+    $neuerwert = $wert + 1;
+        if ($this->class_obj->update("UPDATE pwcounter SET pwcounter=? WHERE id = ? ", array($neuerwert,1))) {
+          $updatedRows = $this->class_obj->rowCount();
+          $this->meldung=" wurde erfolgreich ge&auml;ndert.\n";
+        } else {
+          $this->meldung = "Beim &auml;ndern gab es einen Fehler.\n";
+        }
+      // echo $this->meldung;
+    }
+
   private function update_webcounter() {
     // if ($this->class_obj->update("UPDATE sqlcounter SET sqlcounter=? WHERE id = ? ", array($neuerwert,0))) {
     //   $updatedRows = $this->class_obj->rowCount();
@@ -349,6 +363,12 @@ class PM {
     return $this->sqlcounter ;
   }
 
+  public function pwcounter() {  // 0 = counter, 1 = sqlcounter
+    $counter = $this->class_obj->select("SELECT pwcounter FROM pwcounter WHERE id='1'");
+    $this->pwcounter = $counter[0]['pwcounter'];
+    return $this->pwcounter ;
+  }
+
   public function webcounter() {  // 0 = counter, 1 = sqlcounter
     $counter = $this->class_obj->select("SELECT counter FROM counter WHERE id='1'");
     $this->counter = $counter[0]['counter'];
