@@ -62,11 +62,11 @@ extern "C" {
 	Negative offsets are wrapped to the end of the array (or object) meaning
 	that [-1] references the last item.
 
-	Subscripts may reference a range of offsets by providing 2 offsets seperated
+	Subscripts may reference a range of offsets by providing 2 offsets separated
 	by a colon:
 		foo[2:5]
 
-	Subscripts may reference a set of items by seperating offsets, offset,
+	Subscripts may reference a set of items by separating offsets, offset,
 	ranges, double quoted and single quoted values:
 		foo[2,4:6,'bar.*', "smeg"]
 
@@ -110,10 +110,10 @@ extern "C" {
 		foo; bar <= 3
 		foo; bar != 'foo*'
 		foo; bar != "one"
-
 */
 
-typedef struct WJElementPublic {
+typedef struct WJElementPublic
+{
 	char							*name;
 	WJRType							type;
 
@@ -121,6 +121,7 @@ typedef struct WJElementPublic {
 	struct WJElementPublic			*prev;
 
 	struct WJElementPublic			*child;
+	struct WJElementPublic			*last;
 	struct WJElementPublic			*parent;
 
 	/* The number of children */
@@ -146,9 +147,41 @@ typedef struct WJElementPublic {
 		does get free'd correctly at the correct time.
 	*/
 	XplBool							(* freecb)(struct WJElementPublic *);
+
+	/*
+		If set then this callback will be called when this element is written
+		with WJEWriteDocument() and the callback is responsible for writing the
+		JSON for this node and it's children.
+	*/
+	XplBool							(* writecb)(struct WJElementPublic *, WJWriter writer, char *name);
 } WJElementPublic;
 typedef WJElementPublic *			WJElement;
 
+/*
+	Parse the provided JSON document and return the new WJElement
+
+	The _WJEParse version can be used to parse a document with a non-standard
+	quote character. This allows easy parsing of simple documents directly in a
+	C source file without having to escape double quote characters. Example:
+		doc = _WJEParse("{ 'foo': true, 'bar': 'yup' }", '\'');
+*/
+#define				WJEFromString(j)		__WJEFromString((j), '"', __FILE__, __LINE__)
+#define				WJEParse(j)				__WJEFromString((j), '"', __FILE__, __LINE__)
+#define				_WJEFromString(j, q)	__WJEFromString((j), (q), __FILE__, __LINE__)
+#define				_WJEParse(j, q)			__WJEFromString((j), (q), __FILE__, __LINE__)
+EXPORT WJElement	__WJEFromString(const char *json, char quote, const char *file, const int line);
+
+/*
+	Allocate a string and write the JSON source for the provided WJElement to
+	it. The consumer is responsible for calling MemFree() on the result.
+*/
+EXPORT char *		_WJEToString(WJElement document, XplBool pretty, const char *file, const int line);
+#define WJEToString( d, p ) _WJEToString( (d), (p), __FILE__, __LINE__)
+
+/* Read or write a WJElement to a file by path */
+EXPORT WJElement	WJEFromFile(const char *path);
+EXPORT XplBool		WJEToFile(WJElement document, XplBool pretty, const char *path);
+
 /*
 	Load a WJElement object from the provided WJReader
 
@@ -165,17 +198,29 @@ typedef XplBool		(* WJEWriteCB)(WJElement node, WJWriter writer, void *data);
 EXPORT XplBool		_WJEWriteDocument(WJElement document, WJWriter writer, char *name,
 						WJEWriteCB precb, WJEWriteCB postcb, void *data);
 #define				WJEWriteDocument(d, w, n) _WJEWriteDocument((d), (w), (n), NULL, NULL, NULL)
+
 /* Write a WJElement object to the provided FILE* */
-EXPORT void WJEWriteFILE(WJElement document, FILE* fd);
+EXPORT void			WJEWriteFILE(WJElement document, FILE* fd);
+
+/* Read a WJElement object from the provided FILE* */
+EXPORT WJElement	WJEReadFILE(FILE* fd);
 
 /* Destroy a WJElement object */
-EXPORT XplBool		WJECloseDocument(WJElement document);
+EXPORT XplBool		_WJECloseDocument(WJElement document, const char *file, const int line);
+#define				WJECloseDocument(d) _WJECloseDocument((d), __FILE__, __LINE__)
+
 /*
 	WJECloseDocument is also used to delete/remove an item from a parent
 	document:
 	WJECloseDocument(WJEGet(...));
 */
 
+/*
+	Remove any items matching the specified selector
+*/
+#define				WJERemove(d, p) \
+					while ((WJECloseDocument(WJEGet((d), (p), NULL)))) { ; }
+
 /* Duplicate an existing WJElement */
 typedef XplBool		(* WJECopyCB)(WJElement destination, WJElement object, void *data, const char *file, const int line);
 EXPORT WJElement	_WJECopyDocument(WJElement to, WJElement from, WJECopyCB loadcb, void *data, const char *file, const int line);
@@ -185,6 +230,7 @@ EXPORT WJElement	_WJECopyDocument(WJElement to, WJElement from, WJECopyCB loadcb
 /* Remove a WJElement from it's parent (and siblings) */
 EXPORT XplBool		_WJEDetach(WJElement document, const char *file, const int line);
 #define WJEDetach( d )	_WJEDetach( (d), __FILE__, __LINE__ )
+#define WJEDettach( d )	_WJEDetach( (d), __FILE__, __LINE__ )
 
 /* Add a document to another document as a child */
 EXPORT XplBool		WJEAttach(WJElement container, WJElement document);
@@ -222,7 +268,7 @@ typedef enum {
 		When applicable a NULL will be returned.  Otherwise a value that does
 		not match the requested value will be returned.
 	*/
-	WJE_SET,
+	WJE_SET = 1,
 
 	/*
 		Create a new element and assign it the specified value.
@@ -230,7 +276,7 @@ typedef enum {
 		If an element already exists then it will not be modified, and the value
 		of that existing element will be returned.
 	*/
-	WJE_NEW,
+	WJE_NEW = 2,
 
 	/*
 		Assign the specified value to an existing element, and return the value
@@ -240,9 +286,16 @@ typedef enum {
 		When applicable a NULL will be returned.  Otherwise a value that does
 		not match the requested value will be returned.
 	*/
-	WJE_MOD
+	WJE_MOD = 3
 } WJEAction;
 
+/*
+	The following flags can be OR'ed with a WJEAction value for any function
+	that takes a WJEAction:
+*/
+#define WJE_IGNORE_CASE			0x00010000
+
+
 // TODO	Remove this.  WJE_PUT was renamed to WJE_MOD.
 #define WJE_PUT WJE_MOD
 
@@ -272,16 +325,38 @@ typedef enum {
 #define  WJEDouble( container, path, action,		value)		__WJEDouble(	(container), (path), (action), NULL,	(value),		__FILE__, __LINE__)
 
 EXPORT XplBool		__WJEBool(		WJElement container, const char *path, WJEAction action, WJElement *last, XplBool value,				const char *file, const int line);
-EXPORT char *		__WJEString(	WJElement container, const char *path, WJEAction action, WJElement *last, const char *value,				const char *file, const int line);
-EXPORT char *		__WJEStringN(	WJElement container, const char *path, WJEAction action, WJElement *last, const char *value, size_t len,	const char *file, const int line);
+EXPORT char *		__WJEString(	WJElement container, const char *path, WJEAction action, WJElement *last, const char *value,			const char *file, const int line);
+EXPORT char *		__WJEStringN(	WJElement container, const char *path, WJEAction action, WJElement *last, const char *value, size_t len,const char *file, const int line);
 EXPORT WJElement	__WJEObject(	WJElement container, const char *path, WJEAction action, WJElement *last,								const char *file, const int line);
 EXPORT WJElement	__WJEArray(		WJElement container, const char *path, WJEAction action, WJElement *last,								const char *file, const int line);
 EXPORT WJElement	__WJENull(		WJElement container, const char *path, WJEAction action, WJElement *last,								const char *file, const int line);
-EXPORT int32		__WJEInt32(		WJElement container, const char *path, WJEAction action, WJElement *last, int32 value,				const char *file, const int line);
-EXPORT uint32		__WJEUInt32(	WJElement container, const char *path, WJEAction action, WJElement *last, uint32 value,				const char *file, const int line);
-EXPORT int64		__WJEInt64(		WJElement container, const char *path, WJEAction action, WJElement *last, int64 value,				const char *file, const int line);
-EXPORT uint64		__WJEUInt64(	WJElement container, const char *path, WJEAction action, WJElement *last, uint64 value,				const char *file, const int line);
-EXPORT double		__WJEDouble(	WJElement container, const char *path, WJEAction action, WJElement *last, double value,				const char *file, const int line);
+EXPORT int32		__WJEInt32(		WJElement container, const char *path, WJEAction action, WJElement *last, int32 value,					const char *file, const int line);
+EXPORT uint32		__WJEUInt32(	WJElement container, const char *path, WJEAction action, WJElement *last, uint32 value,					const char *file, const int line);
+EXPORT int64		__WJEInt64(		WJElement container, const char *path, WJEAction action, WJElement *last, int64 value,					const char *file, const int line);
+EXPORT uint64		__WJEUInt64(	WJElement container, const char *path, WJEAction action, WJElement *last, uint64 value,					const char *file, const int line);
+EXPORT double		__WJEDouble(	WJElement container, const char *path, WJEAction action, WJElement *last, double value,					const char *file, const int line);
+
+/*
+	The following functions are identical to the non F variants except that the
+	path argument has been moved to the end and is used as a format string. For
+	example if you need a value by index, and that index is stored in the
+	variable x:
+		val = WJENumberF(doc, WJE_GET, NULL, 0, "foo[%d]", x);
+*/
+EXPORT WJElement	WJEGetF(	WJElement container, WJElement last,													const char *pathf, ...) XplFormatString(3, 4);
+EXPORT XplBool		WJEBoolF(	WJElement container, WJEAction action, WJElement *last, XplBool value,					const char *pathf, ...) XplFormatString(5, 6);
+EXPORT char *		WJEStringF(	WJElement container, WJEAction action, WJElement *last, const char *value,				const char *pathf, ...) XplFormatString(5, 6);
+EXPORT char *		WJEStringNF(WJElement container, WJEAction action, WJElement *last, const char *value, size_t len,	const char *pathf, ...) XplFormatString(6, 7);
+EXPORT WJElement	WJEObjectF(	WJElement container, WJEAction action, WJElement *last,									const char *pathf, ...) XplFormatString(4, 5);
+EXPORT WJElement	WJEArrayF(	WJElement container, WJEAction action, WJElement *last,									const char *pathf, ...) XplFormatString(4, 5);
+EXPORT WJElement	WJENullF(	WJElement container, WJEAction action, WJElement *last,									const char *pathf, ...) XplFormatString(4, 5);
+EXPORT int32		WJEInt32F(	WJElement container, WJEAction action, WJElement *last, int32 value,					const char *pathf, ...) XplFormatString(5, 6);
+EXPORT uint32		WJEUInt32F(	WJElement container, WJEAction action, WJElement *last, uint32 value,					const char *pathf, ...) XplFormatString(5, 6);
+EXPORT int64		WJEInt64F(	WJElement container, WJEAction action, WJElement *last, int64 value,					const char *pathf, ...) XplFormatString(5, 6);
+EXPORT uint64		WJEUInt64F(	WJElement container, WJEAction action, WJElement *last, uint64 value,					const char *pathf, ...) XplFormatString(5, 6);
+EXPORT double		WJEDoubleF(	WJElement container, WJEAction action, WJElement *last, double value,					const char *pathf, ...) XplFormatString(5, 6);
+
+
 
 /*
 	Find, create or update an element by name instead of path.  This allows
@@ -294,6 +369,18 @@ EXPORT double		__WJEDouble(	WJElement container, const char *path, WJEAction act
 EXPORT WJElement _WJEChild(WJElement container, char *name, WJEAction action, const char *file, const int line);
 #define WJEChild(c, n, a) _WJEChild((c), (n), (a), __FILE__, __LINE__)
 
+
+/*
+	Find, create or update an element by path regardless of type.
+
+	Type specific actions may be done by passing the resulting WJElement and a
+	NULL path to WJEBool(), WJENumber(), WJEString(), WJEObject(), WJEArray() or
+	WJENull().
+*/
+EXPORT WJElement _WJEAny(WJElement container, char *path, WJEAction action, WJElement last, const char *file, const int line);
+#define WJEAny(c, p, a, l) _WJEAny((c), (p), (a), (l), __FILE__, __LINE__)
+
+
 /* Calculate a hash for a document */
 typedef int (* WJEHashCB)(void *context, void *data, size_t size);
 EXPORT void WJEHash(WJElement document, WJEHashCB update, void *context);
@@ -372,8 +459,9 @@ EXPORT void WJESchemaFreeBacklink(char *backlink);
 /* Debug function that will write a document to stdout */
 EXPORT void WJEDump(WJElement document);
 
-/* For compatibility with old code, due to a typo */
-#define WJEDettach( d )	_WJEDetach( (d), __FILE__, __LINE__ )
+/* Debug function that will write a document to a file in the current dir */
+EXPORT void WJEDumpFile(WJElement document);
+
 
 #ifdef __cplusplus
 }
