@@ -642,39 +642,154 @@ static XplBool SchemaValidate(WJElement schema, WJElement document,
 #endif
 
 		} else if(!stricmp(memb->name, "additionalProperties")) {
-			sub = WJEObject(schema, "properties", WJE_GET);
-			if(sub && document && document->type == WJR_TYPE_OBJECT) {
-				if(memb->type == WJR_TYPE_FALSE) {
-					data = NULL;
-					while((data = WJEGet(document, "[]", data))) {
-						if(!WJEGet(sub, data->name, NULL)) {
-							fail = TRUE;
-							if(err) {
+			if(document && document->type == WJR_TYPE_OBJECT) {
+				data = NULL;
+				fail = TRUE;
+#ifdef HAVE_REGEX_H
+				regex_t preg;
+#endif
+				while((data = WJEGet(document, "[]", data))) {
+					if((sub = WJEObject(schema, "properties", WJE_GET))) {
+						if(memb->type == WJR_TYPE_FALSE &&
+						   WJEGet(sub, data->name, NULL)) {
+							/* found in properties */
+							if(fail) fail = FALSE;
+						} else if(memb->type == WJR_TYPE_OBJECT) {
+							if(SchemaValidate(memb, data, err,
+											  loadcb, freecb, client,
+											  data->name, version)) {
+								if(fail) fail = FALSE;
+							} else if(err) {
 								err(client,
-									"%s: additional property '%s' found.",
+									"%s: extra property '%s' not valid.",
 									name, data->name);
 							}
 						}
 					}
-				} else if(memb->type == WJR_TYPE_OBJECT) {
-					data = NULL;
-					while((data = WJEGet(document, "[]", data))) {
-						if(!WJEChild(sub, data->name, WJE_GET)) {
-							if(!SchemaValidate(memb, data, err, loadcb, freecb,
-											   client, data->name, version)) {
-								fail = TRUE;
-								if(err) {
-									err(client,
-										"%s: extra property '%s' not valid.",
-										name, data->name);
+#ifdef HAVE_REGEX_H
+					if((sub = WJEObject(schema, "patternProperties",
+										WJE_GET))) {
+						while((arr = WJEGet(sub, "[]", arr))) {
+							if(!regcomp(&preg, arr->name,
+										REG_EXTENDED | REG_NOSUB)) {
+								if(memb->type == WJR_TYPE_FALSE &&
+								   !regexec(&preg, data->name, 0, NULL, 0)) {
+									/* found in patternProperties */
+									if(fail) fail = FALSE;
+								} else if(memb->type == WJR_TYPE_OBJECT) {
+									if(SchemaValidate(memb, data, err,
+													  loadcb, freecb,
+													  client, data->name,
+													  version)) {
+										if(fail) fail = FALSE;
+									} else if(err) {
+										err(client,
+											"%s: extra property '%s' not valid.",
+											name, data->name);
+									}
 								}
+								regfree(&preg);
 							}
 						}
 					}
+#endif
+					if(err && fail) {
+						err(client,
+							"%s: extra property '%s' found.", name, data->name);
+					}
 				}
 				anyFail = anyFail || fail;
 			}
 
+		} else if (!stricmp(memb->name, "anyOf")) {
+			if (document && memb->type == WJR_TYPE_ARRAY) {
+				num = 0;
+				val = 0;
+				arr = NULL;
+				while ((arr = WJEGet(memb, "[]", arr))) {
+					data = NULL;
+					if (val < memb->count) {
+						MemAsprintf(&str, "[%d]", val);
+						data = WJEGet(memb, str, NULL);
+						MemFree(str);
+					}
+					MemAsprintf(&str, "%s[%d]", name, val);
+					if (SchemaValidate(data, document, err, loadcb, freecb,
+						client, str, version)) {
+						num++;
+					}
+					MemFree(str);
+					str = NULL;
+					val++;
+				}
+				if (num == 0) {
+					fail = TRUE;
+					if (err) {
+						err(client,
+							"%s failed anyOf validation.",
+							document->name);
+					}
+				}
+			}
+			anyFail = anyFail || fail;
+
+		} else if (!stricmp(memb->name, "oneOf")) {
+			if (document && memb->type == WJR_TYPE_ARRAY) {
+				num = 0;
+				fail = TRUE;
+				while((arr = WJEGet(memb, "[]", arr))) {
+					if(SchemaValidate(arr, document, err, loadcb, freecb, client, name, version)) {
+						++num;
+						if(num > 1) {
+							break;
+						}
+					}
+				}
+				if(num == 1) {
+					fail = FALSE;
+				}
+				else {
+					if(err) {
+						err(client, "%s did not match exactly one specified schema", name);
+					}
+				}
+				anyFail = anyFail || fail;
+			}
+
+		} else if (!stricmp(memb->name, "allOf")) {
+			if (document && memb->type == WJR_TYPE_ARRAY) {
+				num = 0;
+				val = 0;
+				arr = NULL;
+				while ((arr = WJEGet(memb, "[]", arr))) {
+					data = NULL;
+					if (val < memb->count) {
+						MemAsprintf(&str, "[%d]", val);
+						data = WJEGet(memb, str, NULL);
+						MemFree(str);
+					}
+					MemAsprintf(&str, "%s[%d]", name, val);
+					if (SchemaValidate(data, document, err, loadcb, freecb,
+						client, str, version)) {
+						num++;
+					}
+					MemFree(str);
+					str = NULL;
+					val++;
+				}
+				if (num != memb->count) {
+					fail = TRUE;
+					if (err) {
+						err(client,
+							"%s failed allOf validation; matches: %d, expected: %d", 
+							document->name,
+							num,
+							memb->count);
+					}
+				}
+			}
+			anyFail = anyFail || fail;
+
 		} else if(!stricmp(memb->name, "items")) {
 			if(document && document->type == WJR_TYPE_ARRAY) {
 				if(memb->type == WJR_TYPE_OBJECT) {
@@ -686,7 +801,7 @@ static XplBool SchemaValidate(WJElement schema, WJElement document,
 										   client, str, version)) {
 							fail = TRUE;
 							if(err) {
-								err(client, "%s failed vaidation.", str);
+								err(client, "%s failed validation.", str);
 							}
 						}
 						MemFree(str);
@@ -709,7 +824,7 @@ static XplBool SchemaValidate(WJElement schema, WJElement document,
 										   client, str, version)) {
 							fail = TRUE;
 							if(err) {
-								err(client, "%s failed tuple type vaidation.",
+								err(client, "%s failed tuple type validation.",
 									str);
 							}
 						}
@@ -740,7 +855,7 @@ static XplBool SchemaValidate(WJElement schema, WJElement document,
 							fail = TRUE;
 							if(err) {
 								err(client,
-									"additional item %s failed vaidation.",
+									"additional item %s failed validation.",
 									str);
 							}
 						}
@@ -763,20 +878,18 @@ static XplBool SchemaValidate(WJElement schema, WJElement document,
 				if(fail && err) {
 					err(client, "required item '%s' not found.", name);
 				}
-			} else if(version >= 4 && memb->type == WJR_TYPE_ARRAY) {
+			} else if(memb->type == WJR_TYPE_ARRAY) {
 				/* draft 4;  "required": [ "prop1", "prop2" ] */
 				arr = NULL;
 				while((arr = WJEGet(memb, "[]", arr))) {
-					MemAsprintf(&str, "[\"%s\"]",
-								WJEString(arr, NULL, WJE_GET, ""));
-					if(!document || !WJEGet(document, str, NULL)) {
+					str = WJEString(arr, NULL, WJE_GET, "");
+					if(document && !WJEChild(document, str, WJE_GET)) {
 						fail = TRUE;
 						if(fail && err) {
 							err(client, "%s: required member '%s' not found.",
 								name, str);
 						}
 					}
-					MemRelease(&str);
 				}
 			}
 			anyFail = anyFail || fail;
@@ -828,7 +941,7 @@ static XplBool SchemaValidate(WJElement schema, WJElement document,
 				dval = WJEDouble(memb, NULL, WJE_GET, 0);
 				dnum = WJEDouble(document, NULL, WJE_GET, 0);
 				if(WJEBool(schema, "exclusiveMinimum", WJE_GET, FALSE)) {
-					fail = (dnum < dval && dnum != dval);
+					fail = (dnum < dval || dnum == dval);
 				} else {
 					fail = (dnum < dval);
 				}
@@ -845,7 +958,7 @@ static XplBool SchemaValidate(WJElement schema, WJElement document,
 				dval = WJEDouble(memb, NULL, WJE_GET, 0);
 				dnum = WJEDouble(document, NULL, WJE_GET, 0);
 				if(WJEBool(schema, "exclusiveMaximum", WJE_GET, FALSE)) {
-					fail = (dnum > dval && dnum != dval);
+					fail = (dnum > dval || dnum == dval);
 				} else {
 					fail = (dnum > dval);
 				}
@@ -884,7 +997,40 @@ static XplBool SchemaValidate(WJElement schema, WJElement document,
 				anyFail = anyFail || fail;
 			}
 
-		} else if(!stricmp(memb->name, "minLength")) {
+		}
+		else if (!stricmp(memb->name, "minProperties")) {
+			if (document && document->type == WJR_TYPE_OBJECT &&
+				memb->type == WJR_TYPE_NUMBER) {
+				val = WJENumber(memb, NULL, WJE_GET, 0);
+				fail = (document->count < val);
+				if (fail && err) {
+					err(client,
+						"%s: minimum properties (%d) not met (%d).",
+						name,
+						val,
+						document->count);
+				}
+				anyFail = anyFail || fail;
+			}
+
+		}
+		else if (!stricmp(memb->name, "maxProperties")) {
+			if (document && document->type == WJR_TYPE_OBJECT &&
+				memb->type == WJR_TYPE_NUMBER) {
+				val = WJENumber(memb, NULL, WJE_GET, 0);
+				fail = (document->count > val);
+				if (fail && err) {
+					err(client,
+						"%s: maximum properties (%d) exceeded (%d).",
+						name,
+						val,
+						document->count);
+				}
+				anyFail = anyFail || fail;
+			}
+
+		}
+		else if(!stricmp(memb->name, "minLength")) {
 			if(document && document->type == WJR_TYPE_STRING &&
 			   memb->type == WJR_TYPE_NUMBER) {
 				val = WJENumber(memb, NULL, WJE_GET, 0);
