@@ -17,6 +17,10 @@
 
 #include "element.h"
 
+#ifdef _MSC_VER
+#define strtoull _strtoui64
+#endif
+
 EXPORT XplBool __WJEBool(WJElement container, const char *path, WJEAction action, WJElement *last, XplBool value, const char *file, const int line)
 {
 	_WJElement		*e;
@@ -46,7 +50,7 @@ EXPORT XplBool __WJEBool(WJElement container, const char *path, WJEAction action
 				/* fallthrough */
 
 			default:
-				if (WJE_GET != action) {
+				if (WJE_GET != (action & WJE_ACTION_MASK)) {
 					_WJEChanged(e);
 				}
 				break;
@@ -54,7 +58,7 @@ EXPORT XplBool __WJEBool(WJElement container, const char *path, WJEAction action
 	}
 
 	if (last) *last = (WJElement) e;
-	switch (action) {
+	switch ((action & WJE_ACTION_MASK)) {
 		default:
 		case WJE_GET:
 			if (!e) return(value);
@@ -66,13 +70,23 @@ EXPORT XplBool __WJEBool(WJElement container, const char *path, WJEAction action
 					return(e->value.boolean);
 
 				case WJR_TYPE_NUMBER:
+					return(e->value.number.d != 0);
+
+#ifdef WJE_DISTINGUISH_INTEGER_TYPE
+				case WJR_TYPE_INTEGER:
+					return(e->value.number.i != 0);
+#endif
+
+/*
+				case WJR_TYPE_NUMBER:
+				case WJR_TYPE_INTEGER:
 					if (e->value.number.hasDecimalPoint) {
 						return(e->value.number.d != 0);
 					} else {
 						return(e->value.number.i != 0);
 					}
 					break;
-
+*/
 				default:
 				case WJR_TYPE_UNKNOWN:
 				case WJR_TYPE_NULL:
@@ -114,7 +128,7 @@ static void _WJESetNum(void *dest, size_t size, XplBool issigned, uint64 src, Xp
 			if (!negative) {
 				(*((int64 *) dest)) = src;
 			} else {
-				(*((int64 *) dest)) = -src;
+				(*((int64 *) dest)) = -((int64) src);
 			}
 		} else {
 			(*((uint64 *) dest)) = src;
@@ -124,7 +138,7 @@ static void _WJESetNum(void *dest, size_t size, XplBool issigned, uint64 src, Xp
 			if (!negative) {
 				(*((int32 *) dest)) = src;
 			} else {
-				(*((int32 *) dest)) = -src;
+				(*((int32 *) dest)) = -((int64) src);
 			}
 		} else {
 			(*((uint32 *) dest)) = src;
@@ -175,7 +189,7 @@ static void _WJENum(WJElement container, const char *path, WJEAction action, WJE
 {
 	_WJElement		*e;
 	char			*s, *end;
-	int				r;
+	uint64			r;
 	uint64			v;
 	XplBool			negative;
 
@@ -196,6 +210,9 @@ static void _WJENum(WJElement container, const char *path, WJEAction action, WJE
 				break;
 
 			case WJR_TYPE_NUMBER:
+#ifdef WJE_DISTINGUISH_INTEGER_TYPE
+			case WJR_TYPE_INTEGER:
+#endif
 				negative = FALSE;
 
 				if ((e->value.number.i == _WJEGetNum(value, size, issigned, &negative)) &&
@@ -206,7 +223,7 @@ static void _WJENum(WJElement container, const char *path, WJEAction action, WJE
 				/* fallthrough */
 
 			default:
-				if (WJE_GET != action) {
+				if (WJE_GET != (action & WJE_ACTION_MASK)) {
 					_WJEChanged(e);
 				}
 				break;
@@ -214,7 +231,7 @@ static void _WJENum(WJElement container, const char *path, WJEAction action, WJE
 	}
 
 	if (last) *last = (WJElement) e;
-	switch (action) {
+	switch ((action & WJE_ACTION_MASK)) {
 		default:
 		case WJE_GET:
 			if (!e) {
@@ -224,6 +241,9 @@ static void _WJENum(WJElement container, const char *path, WJEAction action, WJE
 
 			switch (e->pub.type) {
 				case WJR_TYPE_NUMBER:
+#ifdef WJE_DISTINGUISH_INTEGER_TYPE
+				case WJR_TYPE_INTEGER:
+#endif
 					_WJESetNum(value, size, issigned, e->value.number.i, e->value.number.negative);
 					return;
 
@@ -240,12 +260,17 @@ static void _WJENum(WJElement container, const char *path, WJEAction action, WJE
 				case WJR_TYPE_STRING:
 					/* Does the string contain a number? */
 					s = skipspace(e->value.string);
-					r = strtol(s, &end, 0);
-// TODO	Check for a -, and use 64 bit functions
-					if (end != s && (!(s = skipspace(end)) || !*s)) {
-						_WJESetNum(value, size, issigned, r, FALSE);
-						return;
-					}
+				        if(s[0] == '-') {
+				           	negative = TRUE;
+				           	++s;
+				        } else {
+				           	negative = FALSE;
+				        }
+				        r = strtoull(s, &end, 0);
+				        if (end != s && (!(s = skipspace(end)) || !*s)) {
+				        	_WJESetNum(value, size, issigned, r, negative);
+				        	return;
+				        }
 					/* fallthrough */
 
 				default:
@@ -262,8 +287,14 @@ static void _WJENum(WJElement container, const char *path, WJEAction action, WJE
 			if ((e = _WJEReset(e, WJR_TYPE_NUMBER))) {
 				e->value.number.negative		= FALSE;
 				e->value.number.hasDecimalPoint	= FALSE;
+
 				e->value.number.i = _WJEGetNum(value, size, issigned, &e->value.number.negative);
 				e->value.number.d = (double) e->value.number.i;
+#ifdef WJE_DISTINGUISH_INTEGER_TYPE
+				e->pub.type = WJR_TYPE_INTEGER;
+#else
+				e->pub.type = WJR_TYPE_NUMBER;
+#endif
 				return;
 			} else {
 				/* Negate the value - It must NOT match the original */
@@ -278,9 +309,9 @@ static void _WJENum(WJElement container, const char *path, WJEAction action, WJE
 
 EXPORT char * __WJEString(WJElement container, const char *path, WJEAction action, WJElement *last, const char *value, const char *file, const int line)
 {
-	size_t		len = 0;
+	size_t		len	= 0;
 
-	switch (action) {
+	switch ((action & WJE_ACTION_MASK)) {
 		default:
 			break;
 
@@ -320,13 +351,13 @@ EXPORT char * __WJEStringN(WJElement container, const char *path, WJEAction acti
 					break;
 				}
 
-				if (e->value.string && value && !strcmp(e->value.string, value)) {
+				if (e->value.string && value && !wstrcmp(e->value.string, value, action)) {
 					break;
 				}
 				/* fallthrough */
 
 			default:
-				if (WJE_GET != action) {
+				if (WJE_GET != (action & WJE_ACTION_MASK)) {
 					_WJEChanged(e);
 				}
 				break;
@@ -334,7 +365,7 @@ EXPORT char * __WJEStringN(WJElement container, const char *path, WJEAction acti
 	}
 
 	if (last) *last = (WJElement) e;
-	switch (action) {
+	switch ((action & WJE_ACTION_MASK)) {
 		default:
 		case WJE_GET:
 			if (!e) return((char *) value);
@@ -358,6 +389,9 @@ EXPORT char * __WJEStringN(WJElement container, const char *path, WJEAction acti
 
 				default:
 				case WJR_TYPE_NUMBER:
+#ifdef WJE_DISTINGUISH_INTEGER_TYPE
+				case WJR_TYPE_INTEGER:
+#endif
 				case WJR_TYPE_UNKNOWN:
 				case WJR_TYPE_NULL:
 				case WJR_TYPE_OBJECT:
@@ -375,7 +409,7 @@ EXPORT char * __WJEStringN(WJElement container, const char *path, WJEAction acti
 					return((e->value.string = NULL));
 				} else {
 					e->value.string = MemMallocWait(len + 1);
-					strprintf(e->value.string, len + 1, NULL, "%.*s", (int) len, value);
+					strncpy(e->value.string, value, len);
 					e->value.string[len] = '\0';
 
 					MemUpdateOwner(e->value.string, file, line);
@@ -403,7 +437,7 @@ EXPORT WJElement __WJEObject(WJElement container, const char *path, WJEAction ac
 	do {
 		a = action;
 		e = WJESearch(container, path, &a, e, file, line);
-	} while (e && action == WJE_GET && e->type != WJR_TYPE_OBJECT);
+	} while (e && (action & WJE_ACTION_MASK) == WJE_GET && e->type != WJR_TYPE_OBJECT);
 
 	if (e) {
 		switch (e->type) {
@@ -419,7 +453,7 @@ EXPORT WJElement __WJEObject(WJElement container, const char *path, WJEAction ac
 				break;
 
 			default:
-				if (WJE_GET != action) {
+				if (WJE_GET != (action & WJE_ACTION_MASK)) {
 					_WJEChanged(e);
 				}
 				break;
@@ -428,7 +462,7 @@ EXPORT WJElement __WJEObject(WJElement container, const char *path, WJEAction ac
 
 	if (last) *last = e;
 
-	switch (a) {
+	switch ((a & WJE_ACTION_MASK)) {
 		case WJE_SET:
 		case WJE_PUT:
 			return(WJEReset(e, WJR_TYPE_OBJECT));
@@ -455,7 +489,7 @@ EXPORT WJElement __WJEArray(WJElement container, const char *path, WJEAction act
 	do {
 		a = action;
 		e = WJESearch(container, path, &a, e, file, line);
-	} while (e && action == WJE_GET && e->type != WJR_TYPE_ARRAY);
+	} while (e && (action & WJE_ACTION_MASK) == WJE_GET && e->type != WJR_TYPE_ARRAY);
 
 	if (e) {
 		switch (e->type) {
@@ -471,7 +505,7 @@ EXPORT WJElement __WJEArray(WJElement container, const char *path, WJEAction act
 				break;
 
 			default:
-				if (WJE_GET != action) {
+				if (WJE_GET != (action & WJE_ACTION_MASK)) {
 					_WJEChanged(e);
 				}
 				break;
@@ -480,7 +514,7 @@ EXPORT WJElement __WJEArray(WJElement container, const char *path, WJEAction act
 
 	if (last) *last = e;
 
-	switch (a) {
+	switch ((a & WJE_ACTION_MASK)) {
 		case WJE_SET:
 		case WJE_PUT:
 			return(WJEReset(e, WJR_TYPE_ARRAY));
@@ -507,7 +541,7 @@ EXPORT WJElement __WJENull(WJElement container, const char *path, WJEAction acti
 	do {
 		a = action;
 		e = WJESearch(container, path, &a, e, file, line);
-	} while (e && action == WJE_GET && e->type != WJR_TYPE_NULL);
+	} while (e && (action & WJE_ACTION_MASK) == WJE_GET && e->type != WJR_TYPE_NULL);
 
 	if (e) {
 		switch (e->type) {
@@ -523,7 +557,7 @@ EXPORT WJElement __WJENull(WJElement container, const char *path, WJEAction acti
 				break;
 
 			default:
-				if (WJE_GET != action) {
+				if (WJE_GET != (action & WJE_ACTION_MASK)) {
 					_WJEChanged(e);
 				}
 				break;
@@ -532,7 +566,7 @@ EXPORT WJElement __WJENull(WJElement container, const char *path, WJEAction acti
 
 	if (last) *last = e;
 
-	switch (a) {
+	switch ((a & WJE_ACTION_MASK)) {
 		case WJE_SET:
 		case WJE_PUT:
 			return(WJEReset(e, WJR_TYPE_NULL));
@@ -592,7 +626,11 @@ EXPORT double __WJEDouble(WJElement container, const char *path, WJEAction actio
 				break;
 
 			case WJR_TYPE_NUMBER:
-				if (e->value.number.hasDecimalPoint) {
+#ifdef WJE_DISTINGUISH_INTEGER_TYPE
+			case WJR_TYPE_INTEGER:
+#endif
+				if (e->pub.type	== WJR_TYPE_NUMBER)
+				{		
 					if (e->value.number.negative) {
 						if (e->value.number.d * -1.0 == value) {
 							break;
@@ -602,7 +640,9 @@ EXPORT double __WJEDouble(WJElement container, const char *path, WJEAction actio
 							break;
 						}
 					}
-				} else {
+				}
+				else
+				{
 					if (e->value.number.negative) {
 						if ((double) e->value.number.i * -1.0 == value) {
 							break;
@@ -616,7 +656,7 @@ EXPORT double __WJEDouble(WJElement container, const char *path, WJEAction actio
 				/* fallthrough */
 
 			default:
-				if (WJE_GET != action) {
+				if (WJE_GET != (action & WJE_ACTION_MASK)) {
 					_WJEChanged(e);
 				}
 				break;
@@ -624,28 +664,30 @@ EXPORT double __WJEDouble(WJElement container, const char *path, WJEAction actio
 	}
 
 	if (last) *last = (WJElement) e;
-	switch (action) {
+	switch ((action & WJE_ACTION_MASK)) {
 		default:
 		case WJE_GET:
 			if (!e) return(value);
 
 			switch (e->pub.type) {
 				case WJR_TYPE_NUMBER:
-					if (e->value.number.hasDecimalPoint) {
-						if (e->value.number.negative) {
-							return(e->value.number.d * -1.0);
-						} else {
-							return(e->value.number.d);
-						}
+					if (e->value.number.negative) {
+						return(e->value.number.d * -1.0);
 					} else {
-						if (e->value.number.negative) {
-							return((double) e->value.number.i * -1.0);
-						} else {
-							return((double) e->value.number.i);
-						}
+						return(e->value.number.d);
 					}
 					break;
 
+#ifdef WJE_DISTINGUISH_INTEGER_TYPE
+				case WJR_TYPE_INTEGER:
+					if (e->value.number.negative) {
+						return((double) e->value.number.i * -1.0);
+					} else {
+						return((double) e->value.number.i);
+					}
+					break;
+#endif
+
 				case WJR_TYPE_BOOL:
 				case WJR_TYPE_TRUE:
 				case WJR_TYPE_FALSE:
@@ -657,7 +699,7 @@ EXPORT double __WJEDouble(WJElement container, const char *path, WJEAction actio
 					break;
 
 				default:
-					return(value);
+					return(0);
 			}
 
 		case WJE_SET:
@@ -676,7 +718,13 @@ EXPORT double __WJEDouble(WJElement container, const char *path, WJEAction actio
 
 				if (e->value.number.d != (double) e->value.number.i) {
 					e->value.number.hasDecimalPoint = TRUE;
+					e->pub.type = WJR_TYPE_NUMBER;
 				} else {
+#ifdef WJE_DISTINGUISH_INTEGER_TYPE
+					e->pub.type = WJR_TYPE_INTEGER;
+#else
+					e->pub.type = WJR_TYPE_NUMBER;
+#endif
 					e->value.number.hasDecimalPoint = FALSE;
 				}
 
@@ -689,3 +737,313 @@ EXPORT double __WJEDouble(WJElement container, const char *path, WJEAction actio
 	}
 }
 
+static char * WJEPathF(char *buffer, size_t len, const char *pathf, va_list args)
+{
+	size_t		needed;
+	char		*path;
+	va_list		localargs;
+
+	needed		= len;
+	*buffer		= '\0';
+	path		= buffer;
+
+	/* Use a copy of the args to avoid modifying the consumer's data */
+	va_copy(localargs, args);
+
+	vstrcatf(path, len, &needed, pathf, localargs);
+	if (needed <= len) {
+		/* Success */
+		return(path);
+	}
+
+	/*
+		They needed a larger path than the 1024 we put on the stack.
+		Allocate one for them.
+	*/
+	path		= MemMallocWait(++needed);
+	*path		= '\0';
+
+	/* Use a copy of the args to avoid modifying the consumer's data */
+	va_copy(localargs, args);
+
+	vstrcatf(path, needed, &needed, pathf, localargs);
+
+	return(path);
+}
+
+EXPORT WJElement WJEGetF(WJElement container, WJElement last, const char *pathf, ...)
+{
+	WJElement	ret;
+	va_list		args;
+	char		*path;
+	char		buffer[1024];
+
+	va_start(args, pathf);
+
+	path = buffer;
+	path = WJEPathF(buffer, sizeof(buffer), pathf, args);
+
+	va_end(args);
+
+	ret = _WJEGet(container, path, last, __FILE__, __LINE__);
+
+	if (path != buffer) {
+		MemRelease(&path);
+	}
+
+	return(ret);
+}
+
+EXPORT XplBool WJEBoolF(WJElement container, WJEAction action, WJElement *last, XplBool value, const char *pathf, ...)
+{
+	XplBool		ret;
+	va_list		args;
+	char		*path;
+	char		buffer[1024];
+
+	va_start(args, pathf);
+
+	path = buffer;
+	path = WJEPathF(buffer, sizeof(buffer), pathf, args);
+
+	va_end(args);
+
+	ret = __WJEBool(container, path, action, last, value, __FILE__, __LINE__);
+
+	if (path != buffer) {
+		MemRelease(&path);
+	}
+
+	return(ret);
+}
+
+EXPORT char * WJEStringF(WJElement container, WJEAction action, WJElement *last, const char *value, const char *pathf, ...)
+{
+	char *		ret;
+	va_list		args;
+	char		*path;
+	char		buffer[1024];
+
+	va_start(args, pathf);
+
+	path = buffer;
+	path = WJEPathF(buffer, sizeof(buffer), pathf, args);
+
+	va_end(args);
+
+	ret = __WJEString(container, path, action, last, value, __FILE__, __LINE__);
+
+	if (path != buffer) {
+		MemRelease(&path);
+	}
+
+	return(ret);
+}
+
+EXPORT char * WJEStringNF(WJElement container, WJEAction action, WJElement *last, const char *value, size_t len, const char *pathf, ...)
+{
+	char *		ret;
+	va_list		args;
+	char		*path;
+	char		buffer[1024];
+
+	va_start(args, pathf);
+
+	path = buffer;
+	path = WJEPathF(buffer, sizeof(buffer), pathf, args);
+
+	va_end(args);
+
+	ret = __WJEStringN(container, path, action, last, value, len, __FILE__, __LINE__);
+
+	if (path != buffer) {
+		MemRelease(&path);
+	}
+
+	return(ret);
+}
+
+EXPORT WJElement WJEObjectF(WJElement container, WJEAction action, WJElement *last, const char *pathf, ...)
+{
+	WJElement	ret;
+	va_list		args;
+	char		*path;
+	char		buffer[1024];
+
+	va_start(args, pathf);
+
+	path = buffer;
+	path = WJEPathF(buffer, sizeof(buffer), pathf, args);
+
+	va_end(args);
+
+	ret = __WJEObject(container, path, action, last, __FILE__, __LINE__);
+
+	if (path != buffer) {
+		MemRelease(&path);
+	}
+
+	return(ret);
+}
+
+EXPORT WJElement WJEArrayF(WJElement container, WJEAction action, WJElement *last, const char *pathf, ...)
+{
+	WJElement	ret;
+	va_list		args;
+	char		*path;
+	char		buffer[1024];
+
+	va_start(args, pathf);
+
+	path = buffer;
+	path = WJEPathF(buffer, sizeof(buffer), pathf, args);
+
+	va_end(args);
+
+	ret = __WJEArray(container, path, action, last, __FILE__, __LINE__);
+
+	if (path != buffer) {
+		MemRelease(&path);
+	}
+
+	return(ret);
+}
+
+EXPORT WJElement WJENullF(WJElement container, WJEAction action, WJElement *last, const char *pathf, ...)
+{
+	WJElement	ret;
+	va_list		args;
+	char		*path;
+	char		buffer[1024];
+
+	va_start(args, pathf);
+
+	path = buffer;
+	path = WJEPathF(buffer, sizeof(buffer), pathf, args);
+
+	va_end(args);
+
+	ret = __WJENull(container, path, action, last, __FILE__, __LINE__);
+
+	if (path != buffer) {
+		MemRelease(&path);
+	}
+
+	return(ret);
+}
+
+EXPORT int32 WJEInt32F(WJElement container, WJEAction action, WJElement *last, int32 value, const char *pathf, ...)
+{
+	int32		ret;
+	va_list		args;
+	char		*path;
+	char		buffer[1024];
+
+	va_start(args, pathf);
+
+	path = buffer;
+	path = WJEPathF(buffer, sizeof(buffer), pathf, args);
+
+	va_end(args);
+
+	ret = __WJEInt32(container, path, action, last, value, __FILE__, __LINE__);
+
+	if (path != buffer) {
+		MemRelease(&path);
+	}
+
+	return(ret);
+}
+
+EXPORT uint32 WJEUInt32F(WJElement container, WJEAction action, WJElement *last, uint32 value, const char *pathf, ...)
+{
+	uint32		ret;
+	va_list		args;
+	char		*path;
+	char		buffer[1024];
+
+	va_start(args, pathf);
+
+	path = buffer;
+	path = WJEPathF(buffer, sizeof(buffer), pathf, args);
+
+	va_end(args);
+
+	ret = __WJEUInt32(container, path, action, last, value, __FILE__, __LINE__);
+
+	if (path != buffer) {
+		MemRelease(&path);
+	}
+
+	return(ret);
+}
+
+EXPORT int64 WJEInt64F(WJElement container, WJEAction action, WJElement *last, int64 value, const char *pathf, ...)
+{
+	int64		ret;
+	va_list		args;
+	char		*path;
+	char		buffer[1024];
+
+	va_start(args, pathf);
+
+	path = buffer;
+	path = WJEPathF(buffer, sizeof(buffer), pathf, args);
+
+	va_end(args);
+
+	ret = __WJEInt64(container, path, action, last, value, __FILE__, __LINE__);
+
+	if (path != buffer) {
+		MemRelease(&path);
+	}
+
+	return(ret);
+}
+
+EXPORT uint64 WJEUInt64F(WJElement container, WJEAction action, WJElement *last, uint64 value, const char *pathf, ...)
+{
+	uint64		ret;
+	va_list		args;
+	char		*path;
+	char		buffer[1024];
+
+	va_start(args, pathf);
+
+	path = buffer;
+	path = WJEPathF(buffer, sizeof(buffer), pathf, args);
+
+	va_end(args);
+
+	ret = __WJEUInt64(container, path, action, last, value, __FILE__, __LINE__);
+
+	if (path != buffer) {
+		MemRelease(&path);
+	}
+
+	return(ret);
+}
+
+EXPORT double WJEDoubleF(WJElement container, WJEAction action, WJElement *last, double value, const char *pathf, ...)
+{
+	double		ret;
+	va_list		args;
+	char		*path;
+	char		buffer[1024];
+
+	va_start(args, pathf);
+
+	path = buffer;
+	path = WJEPathF(buffer, sizeof(buffer), pathf, args);
+
+	va_end(args);
+
+	ret = __WJEDouble(container, path, action, last, value, __FILE__, __LINE__);
+
+	if (path != buffer) {
+		MemRelease(&path);
+	}
+
+	return(ret);
+}
+
