@@ -20,7 +20,7 @@
 #include <stdlib.h>
 
 #include <xpl.h>
-#include <hulautil.h>
+#include <nmutil.h>
 #include <memmgr.h>
 
 #include <wjreader.h>
@@ -554,13 +554,24 @@ EXPORT size_t WJRFileCallback(char *buffer, size_t length, size_t seen, void *da
 EXPORT size_t WJRMemCallback(char *buffer, size_t length, size_t seen, void *userdata)
 {
 	char		*json = (char *) userdata;
+	char		*end;
 	size_t		len;
 
 	if (!json) {
 		return(0);
 	}
 
+#if 0
 	len = strlen(json);
+#else
+	end = memchr(json + seen, '\0', length + 1);
+	if (end) {
+		len = end - json;
+	} else {
+		len = length + seen;
+	}
+#endif
+
 	if (len <= seen) {
 		return(0);
 	}
